import torch
import torch.nn as nn
from axial_attention import AxialAttention, AxialPositionalEmbedding
from huggingface_hub import PyTorchModelHubMixin

from metnet.layers import ConditionTime, ConvGRU, DownSampler, MetNetPreprocessor, TimeDistributed


class MetNet(torch.nn.Module, PyTorchModelHubMixin):
    def __init__(
        self,
        image_encoder: str = "downsampler",
        input_channels: int = 12,
        sat_channels: int = 12,
        input_size: int = 256,
        output_channels: int = 12,
        hidden_dim: int = 2048,
        kernel_size: int = 3,
        num_layers: int = 1,
        num_att_layers: int = 2,
        num_att_heads: int = 16,
        forecast_steps: int = 48,
        temporal_dropout: float = 0.2,
        use_preprocessor: bool = True,
        **kwargs,
    ):
        super(MetNet, self).__init__()
        config = locals()
        config.pop("self")
        config.pop("__class__")
        self.config = kwargs.pop("config", config)
        sat_channels = self.config["sat_channels"]
        input_size = self.config["input_size"]
        input_channels = self.config["input_channels"]
        temporal_dropout = self.config["temporal_dropout"]
        image_encoder = self.config["image_encoder"]
        forecast_steps = self.config["forecast_steps"]
        hidden_dim = self.config["hidden_dim"]
        kernel_size = self.config["kernel_size"]
        num_layers = self.config["num_layers"]
        num_att_layers = self.config["num_att_layers"]
        output_channels = self.config["output_channels"]
        use_preprocessor = self.config["use_preprocessor"]
        num_att_heads = self.config.get("num_att_heads", 16)

        self.forecast_steps = forecast_steps
        self.input_channels = input_channels
        self.output_channels = output_channels

        if use_preprocessor:
            self.preprocessor = MetNetPreprocessor(
                sat_channels=sat_channels,
                crop_size=input_size,
                use_space2depth=True,
                split_input=True,
            )
            # Update number of input_channels with output from MetNetPreprocessor
            new_channels = sat_channels * 4  # Space2Depth
            new_channels *= 2  # Concatenate two of them together
            input_channels = input_channels - sat_channels + new_channels
        else:
            self.preprocessor = torch.nn.Identity()

        self.drop = nn.Dropout(temporal_dropout)
        if image_encoder in ["downsampler", "default"]:
            image_encoder = DownSampler(input_channels + forecast_steps)
        else:
            raise ValueError(f"Image_encoder {image_encoder} is not recognized")
        self.image_encoder = TimeDistributed(image_encoder)
        self.ct = ConditionTime(forecast_steps)
        self.temporal_enc = TemporalEncoder(
            image_encoder.output_channels, hidden_dim, ks=kernel_size, n_layers=num_layers
        )
        self.position_embedding = AxialPositionalEmbedding(
            dim=self.temporal_enc.out_channels, shape=(input_size // 4, input_size // 4)
        )
        self.temporal_agg = nn.Sequential(
            *[
                AxialAttention(dim=hidden_dim, dim_index=1, heads=num_att_heads, num_dimensions=2)
                for _ in range(num_att_layers)
            ]
        )

        self.head = nn.Conv2d(hidden_dim, output_channels, kernel_size=(1, 1))  # Reduces to mask

    def encode_timestep(self, x, fstep=1):
        # Preprocess Tensor
        x = self.preprocessor(x)

        # Condition Time
        x = self.ct(x, fstep)

        ##CNN
        x = self.image_encoder(x)

        # Temporal Encoder
        _, state = self.temporal_enc(self.drop(x))
        return self.temporal_agg(self.position_embedding(state))

    def forward(self, imgs: torch.Tensor, lead_time: int = 0) -> torch.Tensor:
        """It takes a rank 5 tensor
        - imgs [bs, seq_len, channels, h, w]
        """
        x_i = self.encode_timestep(imgs, lead_time)
        res = self.head(x_i)
        return res


class TemporalEncoder(nn.Module):
    def __init__(self, in_channels, out_channels=384, ks=3, n_layers=1):
        super().__init__()
        self.out_channels = out_channels
        self.rnn = ConvGRU(in_channels, out_channels, (ks, ks), n_layers, batch_first=True)

    def forward(self, x):
        x, h = self.rnn(x)
        return (x, h[-1])


def feat2image(x, target_size=(128, 128)):
    "This idea comes from MetNet"
    x = x.transpose(1, 2)
    return x.unsqueeze(-1).unsqueeze(-1) * x.new_ones(1, 1, 1, *target_size)



"""
Squeeze-and-Excitation net implementation
"""
from typing import Type

import torch
from torch import nn


class SqueezeExcite(nn.Module):
    """
    Squeeze-and-Excitation net
    """

    def __init__(
        self,
        in_channels: int,
        rd_ratio: float = 0.25,
        act_layer: Type[nn.Module] = nn.ReLU,
        gate_layer: Type[nn.Module] = nn.Sigmoid,
    ):
        """
        Constructor Method

        Parameters
        ----------
        in_channels : int
            input channels to layer
        rd_ratio : float, optional
            ratio of squeeze reduction, by default 0.25
        act_layer : Type[nn.Module], optional
            activation layer of containing block, by default nn.ReLU
        gate_layer : Type[nn.Module], optional
            attention gate function, by default nn.Sigmoid
        """
        super().__init__()
        self.in_channels = in_channels
        self.rd_ratio = rd_ratio
        self.out_channels = round(self.in_channels * self.rd_ratio)
        self.conv_reduce = nn.Conv2d(self.in_channels, self.out_channels, 1, bias=True)
        self.act1 = act_layer()
        self.conv_expand = nn.Conv2d(self.out_channels, self.in_channels, 1, bias=True)
        self.gate = gate_layer()

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """
        Forward Pass

        Parameters
        ----------
        X : torch.Tensor
            Input Tensor

        Returns:
        -------
        torch.Tensor
            Output Tensor
        """
        x_se = X.mean((2, 3), keepdim=True)  # Mean along H, W dim
        x_se = self.conv_reduce(x_se)
        x_se = self.act1(x_se)
        x_se = self.conv_expand(x_se)
        return X * self.gate(x_se)



"""
MultiHeaded 2D Self Attention  Implementation
"""
from typing import Type

import torch
import torch.nn as nn


class MultiheadSelfAttention2D(nn.Module):
    """
    Implementing multi-head self-attention for 2D images.
    """

    def __init__(
        self,
        in_channels: int,
        attention_channels: int = 64,
        num_heads: int = 16,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        use_normalised_qk: bool = True,
        rel_attn_bias: Type[nn.Module] = None,
    ) -> None:
        """
        Constructor Method

        Parameters
        ----------
        in_channels : int
            Number of channels in the input image.
        attention_channels : int
            Number of channels used for attention computations.
            It should be divisible by num_heads, by default 64
        num_heads : int
            Number of attention heads, by default 16
        attn_drop : float, optional
            attention dropout rate, by default 0.0
        proj_drop : float, optional
            post attention projection dropout rate, by default 0.0
        use_normalised_qk : bool, by default True
            Normalise queries and keys, (as in Metnet 3)
        rel_attn_bias : Type[nn.Module], optional
            Use Relative Position bias, by default None
        """
        super().__init__()

        assert (
            attention_channels % num_heads == 0
        ), "attention_channels should be divisible by num_heads"

        self.in_channels: int = in_channels
        self.num_heads = num_heads
        self.attention_head_size = attention_channels // num_heads

        # Use relative Position Bias
        self.rel_attn_bias = rel_attn_bias

        # Linear transformations for Q, K and V
        self.query = nn.Conv2d(in_channels, attention_channels, kernel_size=1)
        self.key = nn.Conv2d(in_channels, attention_channels, kernel_size=1)
        self.value = nn.Conv2d(in_channels, attention_channels, kernel_size=1)

        # Output projection
        self.out_proj = nn.Conv2d(attention_channels, in_channels, kernel_size=1)

        # Normalised Keys and Queries as specified in Metnet 3
        self.use_normalised_qk = use_normalised_qk

        self.attn_drop = nn.Dropout(p=attn_drop)
        self.proj_drop = nn.Dropout(p=proj_drop)

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """
        Forward pass for the multi-head self-attention mechanism.

        Parameters
        ----------
        X : torch.Tensor
            Input tensor of shape (N, C, H, W).

        Returns:
        -------
        torch.Tensor
            Output tensor after multi-head self-attention of shape (N, C, H, W).
        """

        N, C, H, W = X.size()
        # Compute Q, K, V
        Q = self.query(X)
        K = self.key(X)
        V = self.value(X)

        Q = Q.view(N, self.num_heads, self.attention_head_size, H * W)
        K = K.view(N, self.num_heads, self.attention_head_size, H * W)
        V = V.view(N, self.num_heads, self.attention_head_size, H * W)

        if self.use_normalised_qk:
            Q = torch.nn.functional.normalize(Q, dim=-2)  # TODO: verify dim
            K = torch.nn.functional.normalize(K, dim=-2)  # TODO: verify dim

        Q = Q.transpose(
            -1, -2
        )  # Making Q of shape [N, self.num_heads, H*W, self.attention_head_size]

        attention_weights = Q @ K  # Attn shape [N, self.num_heads, H*W, H*W]

        if self.rel_attn_bias is not None:
            attention_weights = attention_weights + self.rel_attn_bias()

        attention_weights = attention_weights.softmax(
            dim=-1
        )  # Attn shape [N, self.num_heads, H*W H*W]

        # Multiply attention weights with V
        # V shape [N, self.num_heads, self.attention_head_size, H*W]
        # Attn shape [N, self.num_heads, H*W, H*W]

        attention_weights = self.attn_drop(attention_weights)
        out = V @ attention_weights  # Out shape [N, self.num_heads, H*W, self.attention_head_size]

        # Combine the heads
        out = out.permute(
            0, 1, 3, 2
        )  # Out shape [N, self.num_heads, self.attention_head_size, H*W]
        out = out.contiguous().view(N, self.num_heads * self.attention_head_size, H, W)

        # Apply the output projection
        # Out shape [N, attention_channels, H, W]
        out = self.out_proj(out)  # Out shape [N, C, H, W]
        out = self.proj_drop(out)
        return out + X



"""
MBConv Implementation
"""
from typing import Type

import torch
from torch import nn

from metnet.layers.SqueezeExcitation import SqueezeExcite
from metnet.layers.StochasticDepth import StochasticDepth


class MBConv(nn.Module):
    """
    MB Conv implementation, inspired by timm's implementation.
    """

    def __init__(
        self,
        in_channels: int,
        expansion_rate: int = 4,
        downscale: bool = False,
        act_layer: Type[nn.Module] = nn.GELU,
        norm_layer: Type[nn.Module] = nn.BatchNorm2d,
        drop_path: float = 0.0,
        kernel_size: int = 3,
        se_bottleneck_ratio: float = 0.25,
    ):
        """
        Constructor Method

        Parameters
        ----------
        in_channels : int
            Input channels
        expansion_rate : int, optional
            Expansion rate for the output channels, by default 4
        downscale : bool, optional
            Flag to denote downscaling in the conv branch, by default False
            Currently not implemented, as not specified in Metnet 3
        act_layer : Type[nn.Module], optional
            activation layer, by default nn.GELU
        norm_layer : Type[nn.Module], optional
            normalisation layer, by default nn.BatchNorm2d
            TODO: Verify if Layer Norm is to to be used inside MBConv
            NOTE: Most implementations use nn.BatchNorm2d. If LayerNorm is to be
            used, the intermediate h and w would need to be computed.
        drop_path : float, optional
            Stochastic Depth ratio, by default 0.0
        kernel_size : int, optional
            Conv kernel size, by default 3
        se_bottleneck_ratio : float, optional
            Squeeze Excite reduction ratio, by default 0.25
        """
        # TODO: Verify implemtnetation
        super().__init__()
        self.in_channels = in_channels
        self.drop_path_rate = drop_path
        self.expansion_rate = expansion_rate
        self.downscale = downscale

        out_channels = self.in_channels * self.expansion_rate
        conv_dw_padding = ((kernel_size - 1) // 2, (kernel_size - 1) // 2)

        self.pre_norm = norm_layer(in_channels)
        self.norm_layer_1 = norm_layer(out_channels)
        self.norm_layer_2 = norm_layer(out_channels)

        if self.downscale:
            # TODO: Check if downscaling is needed at all. May impact layer normalisation.
            raise NotImplementedError(
                "Downscaling in MBConv hasn't been implemented as it \
                isnt used in Metnet3"
            )

        self.main_branch = nn.Sequential(
            self.pre_norm,  # Pre Normalize over the last three dimensions (i.e. the channel and spatial dimensions) # noqa
            nn.Conv2d(  # Conv 1x1
                in_channels=self.in_channels,
                out_channels=out_channels,
                kernel_size=1,
            ),
            self.norm_layer_1,  # Norm1
            nn.Conv2d(  # Depth wise Conv kxk
                out_channels,
                out_channels,
                kernel_size,
                padding=conv_dw_padding,  # To maintain shapes
                groups=out_channels,
            ),
            self.norm_layer_2,  # Norm2
            SqueezeExcite(
                in_channels=out_channels,
                act_layer=act_layer,
                rd_ratio=se_bottleneck_ratio,
            ),
            nn.Conv2d(  # Conv 1x1
                in_channels=out_channels,
                out_channels=in_channels,
                kernel_size=1,
            ),
            # No Norm as this is the last convolution layer in this block
        )

        self.stochastic_depth = StochasticDepth(drop_path) if drop_path > 0.0 else nn.Identity()

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """
        Forward step

        Parameters
        ----------
        X : torch.Tensor
            Input Tensor of shape [N, C, H, W]

        Returns:
        -------
        torch.Tensor
            MBConv output
        """
        return X + self.stochastic_depth(self.main_branch(X))



import torch
import torch.nn as nn
import torch.nn.functional as F


class ConvGRUCell(nn.Module):
    def __init__(
        self,
        input_dim,
        hidden_dim,
        kernel_size=(3, 3),
        bias=True,
        activation=torch.tanh,
        batchnorm=False,
    ):
        """
        Initialize ConvGRU cell.
        Parameters
        ----------
        input_dim: int
            Number of channels of input tensor.
        hidden_dim: int
            Number of channels of hidden state.
        kernel_size: (int, int)
            Size of the convolutional kernel.
        bias: bool
            Whether or not to add the bias.
        """
        super().__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim

        self.kernel_size = (
            kernel_size if isinstance(kernel_size, (tuple, list)) else [kernel_size] * 2
        )
        self.padding = self.kernel_size[0] // 2, self.kernel_size[1] // 2
        self.bias = bias
        self.activation = activation
        self.batchnorm = batchnorm

        self.conv_zr = nn.Conv2d(
            in_channels=self.input_dim + self.hidden_dim,
            out_channels=2 * self.hidden_dim,
            kernel_size=self.kernel_size,
            padding=self.padding,
            bias=self.bias,
        )

        self.conv_h1 = nn.Conv2d(
            in_channels=self.input_dim,
            out_channels=self.hidden_dim,
            kernel_size=self.kernel_size,
            padding=self.padding,
            bias=self.bias,
        )

        self.conv_h2 = nn.Conv2d(
            in_channels=self.hidden_dim,
            out_channels=self.hidden_dim,
            kernel_size=self.kernel_size,
            padding=self.padding,
            bias=self.bias,
        )

        self.reset_parameters()

    def forward(self, input, h_prev=None):
        # init hidden on forward
        if h_prev is None:
            h_prev = self.init_hidden(input)

        combined = torch.cat((input, h_prev), dim=1)  # concatenate along channel axis

        combined_conv = torch.sigmoid(self.conv_zr(combined))

        z, r = torch.split(combined_conv, self.hidden_dim, dim=1)

        h_ = self.activation(self.conv_h1(input) + r * self.conv_h2(h_prev))

        h_cur = (1 - z) * h_ + z * h_prev

        return h_cur

    def init_hidden(self, input):
        bs, ch, h, w = input.shape
        return one_param(self).new_zeros(bs, self.hidden_dim, h, w)

    def reset_parameters(self):
        # self.conv.reset_parameters()
        nn.init.xavier_uniform_(self.conv_zr.weight, gain=nn.init.calculate_gain("tanh"))
        self.conv_zr.bias.data.zero_()
        nn.init.xavier_uniform_(self.conv_h1.weight, gain=nn.init.calculate_gain("tanh"))
        self.conv_h1.bias.data.zero_()
        nn.init.xavier_uniform_(self.conv_h2.weight, gain=nn.init.calculate_gain("tanh"))
        self.conv_h2.bias.data.zero_()

        if self.batchnorm:
            self.bn1.reset_parameters()
            self.bn2.reset_parameters()


def one_param(m):
    "First parameter in `m`"
    return next(m.parameters())


def dropout_mask(x, sz, p):
    "Return a dropout mask of the same type as `x`, size `sz`, with probability `p` to cancel an element."
    return x.new_empty(*sz).bernoulli_(1 - p).div_(1 - p)


class RNNDropout(nn.Module):
    "Dropout with probability `p` that is consistent on the seq_len dimension."

    def __init__(self, p=0.5):
        super().__init__()
        self.p = p

    def forward(self, x):
        if not self.training or self.p == 0.0:
            return x
        return x * dropout_mask(x.data, (x.size(0), 1, *x.shape[2:]), self.p)


class ConvGRU(nn.Module):
    def __init__(
        self,
        input_dim,
        hidden_dim,
        kernel_size,
        n_layers,
        batch_first=True,
        bias=True,
        activation=F.tanh,
        input_p=0.2,
        hidden_p=0.1,
        batchnorm=False,
    ):
        super(ConvGRU, self).__init__()

        self._check_kernel_size_consistency(kernel_size)

        # Make sure that both `kernel_size` and `hidden_dim` are lists having len == num_layers
        kernel_size = self._extend_for_multilayer(kernel_size, n_layers)
        hidden_dim = self._extend_for_multilayer(hidden_dim, n_layers)
        activation = self._extend_for_multilayer(activation, n_layers)

        if not len(kernel_size) == len(hidden_dim) == len(activation) == n_layers:
            raise ValueError("Inconsistent list length.")

        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.kernel_size = kernel_size
        self.n_layers = n_layers
        self.batch_first = batch_first
        self.bias = bias
        self.input_p = input_p
        self.hidden_p = hidden_p

        cell_list = []
        for i in range(self.n_layers):
            cur_input_dim = self.input_dim if i == 0 else self.hidden_dim[i - 1]

            cell_list.append(
                ConvGRUCell(
                    input_dim=cur_input_dim,
                    hidden_dim=self.hidden_dim[i],
                    kernel_size=self.kernel_size[i],
                    bias=self.bias,
                    activation=activation[i],
                    batchnorm=batchnorm,
                )
            )

        self.cell_list = nn.ModuleList(cell_list)
        self.input_dp = RNNDropout(input_p)
        self.hidden_dps = nn.ModuleList([nn.Dropout(hidden_p) for l in range(n_layers)])
        self.reset_parameters()

    def __repr__(self):
        s = f"ConvGru(in={self.input_dim}, out={self.hidden_dim[0]}, ks={self.kernel_size[0]}, "
        s += f"n_layers={self.n_layers}, input_p={self.input_p}, hidden_p={self.hidden_p})"
        return s

    def forward(self, input, hidden_state=None):
        """
        Parameters
        ----------
        input_tensor:
            5-D Tensor either of shape (t, b, c, h, w) or (b, t, c, h, w)
        hidden_state:
        Returns
        -------
        last_state_list, layer_output
        """
        input = self.input_dp(input)
        cur_layer_input = torch.unbind(input, dim=int(self.batch_first))

        if hidden_state is None:
            hidden_state = self.get_init_states(cur_layer_input[0])

        seq_len = len(cur_layer_input)

        last_state_list = []

        for l, (gru_cell, hid_dp) in enumerate(zip(self.cell_list, self.hidden_dps)):
            h = hidden_state[l]
            output_inner = []
            for t in range(seq_len):
                h = gru_cell(input=cur_layer_input[t], h_prev=h)
                output_inner.append(h)

            cur_layer_input = torch.stack(output_inner)  # list to array
            if l != self.n_layers:
                cur_layer_input = hid_dp(cur_layer_input)
            last_state_list.append(h)

        layer_output = torch.stack(output_inner, dim=int(self.batch_first))
        last_state_list = torch.stack(last_state_list, dim=0)
        return layer_output, last_state_list

    def reset_parameters(self):
        for c in self.cell_list:
            c.reset_parameters()

    def get_init_states(self, input):
        init_states = []
        for gru_cell in self.cell_list:
            init_states.append(gru_cell.init_hidden(input))
        return init_states

    @staticmethod
    def _check_kernel_size_consistency(kernel_size):
        if not (
            isinstance(kernel_size, tuple)
            or (
                isinstance(kernel_size, list)
                and all([isinstance(elem, tuple) for elem in kernel_size])
            )
        ):
            raise ValueError("`kernel_size` must be tuple or list of tuples")

    @staticmethod
    def _extend_for_multilayer(param, num_layers):
        if not isinstance(param, list):
            param = [param] * num_layers
        return param



"""
Implementation of Partition (Grid and Block) Attention
"""
from typing import Tuple, Type

import torch
from torch import Tensor, nn

from metnet.layers.MultiheadSelfAttention2D import MultiheadSelfAttention2D
from metnet.layers.RelativePositionBias import RelativePositionBias
from metnet.layers.StochasticDepth import StochasticDepth


class PointwiseMLP(nn.Module):
    """
    Pointwise MLP for [N, C, H, W] images
    """

    def __init__(self, in_channels: int, out_channels: int = None, hidden_dim: int = None) -> None:
        """
        Constructor Method

        Parameters
        ----------
        in_channels : int
            Input Channels
        out_channels : int, optional
            Output Channels, by default None
            If None, set equal to in_channels
        hidden_dim : int, optional
            Hidden Dim for MLP, by default None
            If None, set equal to 4 * in_channels
        """
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels if out_channels else in_channels
        self.hidden_dim = hidden_dim if hidden_dim else 4 * self.in_channels

        self.mlp = nn.Sequential(
            nn.Linear(self.in_channels, self.hidden_dim),
            nn.GELU(),
            nn.Linear(self.hidden_dim, self.out_channels),
        )

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """
        Forward Method

        Parameters
        ----------
        X : torch.Tensor
            Input Tensor

        Returns:
        -------
        torch.Tensor
            Output Tensor
        """
        X = X.permute(0, 2, 3, 1)  # Converting from [N, C, H, W] to [N, W, C, W]
        return self.mlp(X)


class PartitionAttention(nn.Module):
    """
    Partition Attention

    Implements the common functionality for block and grid attention.
    X ← X + StochasticDepth(RelAttention(PreNorm(Partition(X))))
    X ← X + StochasticDepth(MLP(PostNorm(X)))
    X ← ReversePartition(X)

    """

    def __init__(
        self,
        in_channels: int,
        num_heads: int = 32,
        attention_channels: int = 64,
        attn_grid_window_size: Tuple[int, int] = (8, 8),
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        drop_path: float = 0.0,
        pre_norm_layer: Type[nn.Module] = nn.LayerNorm,
        post_norm_layer: Type[nn.Module] = nn.LayerNorm,
        use_mlp: bool = False,
        use_normalised_qk: bool = True,
    ) -> None:
        """
        Constructor Method

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        num_heads : int, optional
             Number of attention heads, by default 32
        attention_channels : int
            Number of channels used for attention computations.
            It should be divisible by num_heads, by default 64
        attn_grid_window_size : Tuple[int, int], optional
             Grid/Window size to be utilized, by default (8, 8)
        attn_drop : float, optional
             Dropout ratio of attention weight, by default 0.0
        proj_drop : float, optional
            Dropout ratio of output, by default 0.0
        drop_path : float, optional
            Stochastic depth, by default 0.0
        pre_norm_layer : Type[nn.Module], optional
            Pre norm layer, by default nn.LayerNorm
        post_norm_layer : Type[nn.Module], optional
            Post norm layer, by default nn.LayerNorm
        use_mlp : bool, optional
            MLP to be used after the attention, by default False
        use_normalised_qk : bool, optional
            Normalise queries and keys as done in Metnet 3, by default True.

        Notes:
        -----
        Specific to Metnet 3 implementation
        """
        super().__init__()
        # Save parameters
        self.attn_grid_window_size: Tuple[int, int] = attn_grid_window_size

        rel_attn_bias = RelativePositionBias(attn_size=attn_grid_window_size, num_heads=num_heads)
        # Init layers
        self.attention = MultiheadSelfAttention2D(
            in_channels=in_channels,
            attention_channels=attention_channels,
            num_heads=num_heads,
            attn_drop=attn_drop,
            proj_drop=proj_drop,
            use_normalised_qk=use_normalised_qk,
            rel_attn_bias=rel_attn_bias,
        )
        self.pre_norm_layer = pre_norm_layer(attn_grid_window_size)  # Norm along windows
        self.post_norm_layer = post_norm_layer(attn_grid_window_size)

        if use_mlp:
            self.mlp = PointwiseMLP(in_channels=in_channels)
        else:
            self.mlp = nn.Identity()
        self.drop_path = StochasticDepth(drop_path) if drop_path > 0.0 else nn.Identity()

    def partition_function(self, X: torch.Tensor):
        """
        Partition function.

        To be overridden by block or grid partition

        Parameters
        ----------
        X : torch.Tensor
            Input tensor

        Raises:
        ------
        NotImplementedError
            Should not be called without implementation in the child class
        """
        raise NotImplementedError

    def reverse_function(
        self,
        partitioned_input: torch.Tensor,
        original_size: Tuple[int, int],
    ):
        """
        Undo Partition

        To be overridden by functions reversing the block or grid partitions

        Parameters
        ----------
        partitioned_input : torch.Tensor
            Partitioned input
        original_size : Tuple[int, int]
            Original Input size

        Raises:
        ------
        NotImplementedError
            Should not be called without implementation in the child class
        """
        raise NotImplementedError

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """
        Forward pass.

        Parameters
        ----------
        X : torch.Tensor
            Input tensor of the shape [N, C, H, W].

        Returns:
        -------
        torch.Tensor
            Output tensor of the shape [N, C, H, W].
        """
        # Save original shape
        _, C, H, W = X.shape
        # Perform partition
        input_partitioned = self.partition_function(X)

        # Perform normalization, attention, and dropout
        output = input_partitioned + self.drop_path(
            self.attention(self.pre_norm_layer(input_partitioned))
        )

        # Perform normalization, MLP, and dropout
        output = output + self.drop_path(self.mlp(self.post_norm_layer(output)))

        # Reverse partition
        output = self.reverse_function(output, (H, W))
        return output


class BlockAttention(PartitionAttention):
    """
    Block Attention.
    """

    def __init__(
        self,
        in_channels: int,
        num_heads: int = 32,
        attention_channels: int = 64,
        attn_grid_window_size: Tuple[int, int] = (8, 8),
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        drop_path: float = 0.0,
        pre_norm_layer: Type[nn.Module] = nn.LayerNorm,
        post_norm_layer: Type[nn.Module] = nn.LayerNorm,
        use_mlp: bool = False,
        use_normalised_qk: bool = True,
    ) -> None:
        """
        Constructor Method

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        num_heads : int, optional
             Number of attention heads, by default 32
        attention_channels : int
            Number of channels used for attention computations.
            It should be divisible by num_heads, by default 64
        attn_grid_window_size : Tuple[int, int], optional
             Grid/Window size for attention, by default (8, 8)
        attn_drop : float, optional
             Dropout ratio of attention weight, by default 0.0
        proj_drop : float, optional
            Dropout ratio of output, by default 0.0
        drop_path : float, optional
            Stochastic depth, by default 0.0
        pre_norm_layer : Type[nn.Module], optional
            Pre norm layer, by default nn.LayerNorm
        post_norm_layer : Type[nn.Module], optional
            Post norm layer, by default nn.LayerNorm
        use_mlp : bool, optional
            MLP to be used after the attention, by default False
        use_normalised_qk : bool, optional
            Normalise queries and keys as done in Metnet 3, by default True.
        """
        super().__init__(
            in_channels,
            num_heads,
            attention_channels,
            attn_grid_window_size,
            attn_drop,
            proj_drop,
            drop_path,
            pre_norm_layer,
            post_norm_layer,
            use_mlp,
            use_normalised_qk,
        )

    def partition_function(self, input: Tensor) -> torch.Tensor:
        """
        Block partition function.

        Parameters
        ----------
        input : Tensor
            input (torch.Tensor): Input tensor of the shape [N, C, H, W].

        Returns:
        -------
        torch.Tensor
            blocks (torch.Tensor): Unfolded input tensor of the shape
            [N * blocks, partition_size[0], partition_size[1], C].
        """
        # Get size of input
        N, C, H, W = input.shape
        # Unfold input
        blocks = input.view(
            N,
            C,
            H // self.attn_grid_window_size[0],
            self.attn_grid_window_size[0],
            W // self.attn_grid_window_size[1],
            self.attn_grid_window_size[1],
        )
        # Permute and reshape to
        # [N * blocks, channels, self.attn_grid_window_size[0], self.attn_grid_window_size[1]]
        blocks = (
            blocks.permute(0, 2, 4, 3, 5, 1)
            .contiguous()
            .view(-1, C, self.attn_grid_window_size[0], self.attn_grid_window_size[1])
        )
        return blocks

    def reverse_function(
        self,
        partitioned_input: torch.Tensor,
        original_size: Tuple[int, int],
    ) -> torch.Tensor:
        """
        Reverses the block partition.

        Parameters
        ----------
        partitioned_input : torch.Tensor
            Block tensor of the shape
            [N * partitioned_input, C, partition_size[0], partition_size[1]].
        original_size : Tuple[int, int]
            Original shape.

        Returns:
        -------
        torch.Tensor
            output (torch.Tensor): Folded output tensor of the shape
            [N, C, original_size[0], original_size[1]].
        """
        # Get height and width
        H, W = original_size
        # Compute original batch size
        N = int(
            partitioned_input.shape[0]
            / (H * W / self.attn_grid_window_size[0] / self.attn_grid_window_size[1])
        )
        # Fold grid tensor
        output = partitioned_input.view(  # TODO: Verify
            N,
            -1,
            H // self.attn_grid_window_size[0],
            W // self.attn_grid_window_size[1],
            self.attn_grid_window_size[0],
            self.attn_grid_window_size[1],
        )
        output = output.contiguous().view(N, -1, H, W)
        return output


class GridAttention(PartitionAttention):
    """
    Grid Attention
    """

    def __init__(
        self,
        in_channels: int,
        num_heads: int = 32,
        attention_channels: int = 64,
        attn_grid_window_size: Tuple[int, int] = (8, 8),
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        drop_path: float = 0.0,
        pre_norm_layer: Type[nn.Module] = nn.LayerNorm,
        post_norm_layer: Type[nn.Module] = nn.LayerNorm,
        use_mlp: bool = False,
        use_normalised_qk: bool = True,
    ) -> None:
        """
        Constructor Method

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        num_heads : int, optional
             Number of attention heads, by default 32
        attention_channels : int
            Number of channels used for attention computations.
            It should be divisible by num_heads, by default 64
        attn_grid_window_size : Tuple[int, int], optional
             Grid/Window size to be utilized, by default (8, 8)
        attn_drop : float, optional
             Dropout ratio of attention weight, by default 0.0
        proj_drop : float, optional
            Dropout ratio of output, by default 0.0
        drop_path : float, optional
            Stochastic depth, by default 0.0
        pre_norm_layer : Type[nn.Module], optional
            Pre norm layer, by default nn.LayerNorm
        post_norm_layer : Type[nn.Module], optional
            Post norm layer, by default nn.LayerNorm
        use_mlp : bool, optional
            MLP to be used after the attention, by default False
        use_normalised_qk : bool, optional
            Normalise queries and keys as done in Metnet 3, by default True.
        """
        super().__init__(
            in_channels,
            num_heads,
            attention_channels,
            attn_grid_window_size,
            attn_drop,
            proj_drop,
            drop_path,
            pre_norm_layer,
            post_norm_layer,
            use_mlp,
            use_normalised_qk,
        )

    def partition_function(self, input: Tensor) -> torch.Tensor:
        """
        Grid partition function.

        Parameters
        ----------
        input : Tensor
            Input tensor of the shape [N, C, H, W].

        Returns:
        -------
        torch.Tensor
            Unfolded input tensor of the shape
            [N * grids, grid_size[0], grid_size[1], C].
        """
        # Get size of input
        N, C, H, W = input.shape
        # Unfold input
        grid = input.view(
            N,
            C,
            self.attn_grid_window_size[0],
            H // self.attn_grid_window_size[0],
            self.attn_grid_window_size[1],
            W // self.attn_grid_window_size[1],
        )
        # Permute and reshape [N * (H // self.attn_grid_window_size[0]) * (W // self.attn_grid_window_size[1]), C, self.attn_grid_window_size[0], self.attn_grid_window_size[1]]  # noqa
        grid = (
            grid.permute(0, 3, 5, 1, 2, 4)
            .contiguous()
            .view(-1, C, self.attn_grid_window_size[0], self.attn_grid_window_size[1])
        )
        return grid

    def reverse_function(
        self,
        partitioned_input: torch.Tensor,
        original_size: Tuple[int, int],
    ) -> torch.Tensor:
        """
        Reverses the grid partition.

        Parameters
        ----------
        partitioned_input : torch.Tensor
            Grid tensor of the shape
            [N * partitioned_input, C, partition_size[0], partition_size[1]].
        original_size : Tuple[int, int]
            Original shape.

        Returns:
        -------
        torch.Tensor
            Folded output tensor of the shape [N, C, original_size[0], original_size[1]].
        """
        # Get height, width, and channels
        (H, W), C = original_size, partitioned_input.shape[1]
        # Compute original batch size
        N = int(
            partitioned_input.shape[0]
            / (H * W / self.attn_grid_window_size[0] / self.attn_grid_window_size[1])
        )
        # Fold partitioned_input tensor
        output = partitioned_input.view(  # TODO: Verify
            N,
            C,
            H // self.attn_grid_window_size[0],
            W // self.attn_grid_window_size[1],
            self.attn_grid_window_size[0],
            self.attn_grid_window_size[1],
        )
        output = output.permute(0, 1, 3, 5, 4, 2).contiguous().view(N, C, H, W)
        return output



"""
Implementation of Stochastic Depth
"""
import torch
from torch import nn


class StochasticDepth(nn.Module):
    """
    Stochastic Depth

    Drops network paths with the given probability
    """

    def __init__(self, drop_prob: float = 0.0) -> None:
        """
        Constructor Method

        Parameters
        ----------
        drop_prob : float, optional
            probability to drop the network path, by default 0.0
        """
        super().__init__()
        assert 0 <= drop_prob <= 1.0
        self.drop_prob = drop_prob

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """
        Forward Pass

        Parameters
        ----------
        X : torch.Tensor
            Input Path

        Returns:
        -------
        torch.Tensor
            Output tensor. Zeroed out with the given probability.
        """
        if torch.rand(1).item() < self.drop_prob:
            X = X * 0
        return X



"""Utilities"""
import numpy as np


def mrms_normalize(a):
    """
    Normalize, from MetNet-2 code release

    [NaN, inf, -inf, -50, -.5, 0., .2, 1., 2., 10.])
    ->
    [-1, -1, -1, -1, -1, 0, .046, .172, .268, .537]
    """
    a = np.nan_to_num(np.where(a < 0, 0, a + 1), posinf=0, neginf=0)
    return np.tanh(np.log(a) / 4)



import torch
import torch.nn.functional as F
from metnet import MetNet, MetNet2, MetNetPV


def test_metnet_creation():
    model = MetNet(
        hidden_dim=32,
        forecast_steps=24,
        input_channels=16,
        output_channels=12,
        sat_channels=12,
        input_size=64,
    )
    # MetNet expects original HxW to be 4x the input size
    x = torch.randn((2, 12, 16, 256, 256))
    model.eval()
    with torch.no_grad():
        out = model(x)
    # MetNet creates predictions for the center 1/4th
    assert out.size() == (
        2,
        12,
        16,
        16,
    )
    assert not torch.isnan(out).any(), "Output included NaNs"


def test_metnet_backwards():
    model = MetNet(
        hidden_dim=32,
        forecast_steps=24,
        input_channels=16,
        output_channels=12,
        sat_channels=12,
        input_size=32,
    )
    # MetNet expects original HxW to be 4x the input size
    x = torch.randn((2, 12, 16, 128, 128))
    out = []
    for lead_time in range(24):
        out.append(model(x, lead_time))
    out = torch.stack(out, dim=1)
    # MetNet creates predictions for the center 1/4th
    assert out.size() == (
        2,
        24,
        12,
        8,
        8,
    )
    y = torch.randn((2, 24, 12, 8, 8))
    F.mse_loss(out, y).backward()
    assert not torch.isnan(out).any(), "Output included NaNs"


def test_metnet_pv_backwards():
    model = MetNetPV(
        hidden_dim=128,
        num_att_layers=2,
        num_att_heads=16,
        forecast_steps=24,
        input_channels=16,
        output_channels=1,
        sat_channels=12,
        input_size=64,
        avg_pool_size=2,
        pv_fc_out_channels=128,
        pv_id_embedding_channels=256,
    )
    # MetNet expects original HxW to be 4x the input size
    x = torch.randn((2, 12, 16, 256, 256))
    pv_x = torch.randn((2, 12, 1, 1000))
    pv_idx = torch.randint(5000, (2, 1000))
    out = []
    for lead_time in range(24):
        out.append(model(x, pv_x, pv_idx, lead_time))
    out = torch.stack(out, dim=1)
    # MetNet creates predictions for the center 1/4th
    assert out.size() == (2, 24, 1)
    y = torch.randn((2, 24, 1))
    F.mse_loss(out, y).backward()
    assert not torch.isnan(out).any(), "Output included NaNs"


def test_load_metnet_hf():
    model = MetNet.from_pretrained("openclimatefix/metnet")


def test_load_metnet2_hf():
    model = MetNet2.from_pretrained("openclimatefix/metnet-2")


def test_metnet2_creation():
    model = MetNet2(
        forecast_steps=8,
        input_size=128,
        num_input_timesteps=6,
        upsampler_channels=128,
        lstm_channels=32,
        encoder_channels=64,
        center_crop_size=32,
    )
    # MetNet expects original HxW to be 4x the input size
    x = torch.randn((2, 6, 12, 512, 512))
    model.eval()
    with torch.no_grad():
        out = model(x)
    # MetNet creates predictions for the center 1/4th
    assert out.size() == (
        2,
        12,
        128,
        128,
    )
    assert not torch.isnan(out).any(), "Output included NaNs"


def test_metnet2_backward():
    model = MetNet2(
        forecast_steps=4,
        input_size=64,
        num_input_timesteps=6,
        upsampler_channels=128,
        lstm_channels=32,
        encoder_channels=64,
        center_crop_size=16,
    )
    # MetNet expects original HxW to be 4x the input size
    x = torch.randn((2, 6, 12, 256, 256))
    out = []
    for lead_time in range(4):
        out.append(model(x, lead_time))
    out = torch.stack(out, dim=1)
    # MetNet creates predictions for the center 1/4th
    assert out.size() == (
        2,
        4,
        12,
        64,
        64,
    )
    y = torch.rand((2, 4, 12, 64, 64))
    F.mse_loss(out, y).backward()
    assert not torch.isnan(out).any(), "Output included NaNs"



from metnet.layers.MaxViT import MaxViTBlock, MaxViTDataClass, MetNetMaxVit
from metnet.layers.StochasticDepth import StochasticDepth
from metnet.layers.SqueezeExcitation import SqueezeExcite
from metnet.layers.MBConv import MBConv
from metnet.layers.MultiheadSelfAttention2D import MultiheadSelfAttention2D
from metnet.layers.PartitionAttention import BlockAttention, GridAttention
import torch


def test_stochastic_depth():
    test_tensor = torch.ones(1)

    stochastic_depth = StochasticDepth(drop_prob=0)
    assert test_tensor == stochastic_depth(test_tensor)

    stochastic_depth = StochasticDepth(drop_prob=1)
    assert torch.zeros_like(test_tensor) == stochastic_depth(test_tensor)


def test_squeeze_excitation():
    n, c, h, w = 1, 3, 16, 16
    test_tensor = torch.rand(n, c, h, w)

    squeeze_excite = SqueezeExcite(in_channels=c)
    assert test_tensor.shape == squeeze_excite(test_tensor).shape


def test_mbconv():
    n, c, h, w = 1, 3, 16, 16
    test_tensor = torch.rand(n, c, h, w)
    mb_conv = MBConv(c)

    assert test_tensor.shape == mb_conv(test_tensor).shape


def test_multiheaded_self_attention_2D():
    n, c, h, w = 1, 3, 16, 16
    test_tensor = torch.rand(n, c, h, w)
    rel_self_attention = MultiheadSelfAttention2D(c)
    assert test_tensor.shape == rel_self_attention(test_tensor).shape


def test_block_attention():
    n, c, h, w = 1, 3, 16, 16
    test_tensor = torch.rand(n, c, h, w)
    block_attention = BlockAttention(c)

    assert test_tensor.shape == block_attention(test_tensor).shape


def test_grid_attention():
    n, c, h, w = 1, 3, 16, 16
    test_tensor = torch.rand(n, c, h, w)
    grid_attention = GridAttention(c)

    assert test_tensor.shape == grid_attention(test_tensor).shape


def test_maxvitblock():
    n, c, h, w = 1, 3, 16, 16
    test_tensor = torch.rand(n, c, h, w)

    maxvit_block = MaxViTBlock(in_channels=c, maxvit_config=MaxViTDataClass())
    assert test_tensor.shape == maxvit_block(test_tensor).shape


def test_metnet_maxvit():
    n, c, h, w = 1, 3, 16, 16
    test_tensor = torch.rand(n, c, h, w)

    metnet_maxvit = MetNetMaxVit(in_channels=c)
    assert test_tensor.shape == metnet_maxvit(test_tensor).shape



import torch
import torch.nn.functional as F
import torchvision
from torch import nn


class MetNetPreprocessor(nn.Module):
    def __init__(
        self,
        sat_channels: int = 12,
        crop_size: int = 256,
        use_space2depth: bool = True,
        split_input: bool = True,
    ):
        """
        Performs the MetNet preprocessing of mean pooling Sat channels, followed by
        concatenating the center crop and mean pool

        In the paper, the radar data is space2depth'd, while satellite channel is mean pooled, but for this different
        task, we choose to do either option for satellites
        Args:
            sat_channels: Number of satellite channels
            crop_size: Center crop size
            use_space2depth: Whether to use space2depth on satellite channels, or mean pooling, like in paper

        """
        super().__init__()
        self.sat_channels = sat_channels
        self.split_input = split_input

        # Split off sat + mask channels into own image, and the rest, which we just take a center crop
        # For this,
        self.sat_downsample = (
            torch.nn.PixelUnshuffle(downscale_factor=2)
            if use_space2depth
            else torch.nn.AvgPool3d(kernel_size=(1, 2, 2))
        )
        self.center_crop = torchvision.transforms.CenterCrop(size=crop_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.split_input:
            sat_channels = x[:, :, : self.sat_channels, :, :]
            other_channels = x[:, :, self.sat_channels :, :, :]
            other_channels = torchvision.transforms.CenterCrop(size=other_channels.size()[-1] // 2)(
                other_channels
            )  # center crop to same as downsample
            other_channels = self.center_crop(other_channels)
        else:
            sat_channels = x
        sat_channels = self.sat_downsample(sat_channels)
        # In paper, satellite and radar data is concatenated here
        # We are just going to skip that bit

        sat_center = self.center_crop(sat_channels)
        sat_mean = F.avg_pool3d(sat_channels, (1, 2, 2))
        # All the same size now, so concatenate together, already have time, lat/long, and elevation image
        x = (
            torch.cat([sat_center, sat_mean, other_channels], dim=2)
            if self.split_input
            else torch.cat([sat_center, sat_mean], dim=2)
        )
        return x



"""
Implementation of Relative Position Bias
"""
from typing import Tuple

import torch
import torch.nn as nn


class RelativePositionBias(nn.Module):
    """
    Relative Postition Bias

    Inspired by timm's maxxvit implementation
    """

    def __init__(self, attn_size: Tuple[int, int], num_heads: int) -> None:
        """
        Constructor Method

        Parameters
        ----------
        attn_size : Tuple[int, int]
            Size of the attention window
        num_heads : int
            Number of heads in the multiheaded attention
        """
        super().__init__()
        self.attn_size = attn_size
        self.attn_area = self.attn_size[0] * self.attn_size[1]
        self.num_heads = num_heads

        # Define a parameter table of relative position bias, shape: 2*Wh-1 * 2*Ww-1, nH
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros(
                (2 * self.attn_size[0] - 1) * (2 * self.attn_size[1] - 1),
                self.num_heads,
            )
        )

        # Get pair-wise relative position index for each token inside the window
        self.register_buffer(
            "relative_position_index",
            self.get_relative_position_index(self.attn_size[0], self.attn_size[1]),
        )

    def get_relative_position_index(self, win_h: int, win_w: int) -> torch.Tensor:
        """
        Function to generate pair-wise relative position index for each token inside the window.

        Taken from Timms Swin V1 implementation.

        Parameters
        ----------
        win_h : int
            Window/Grid height.
        win_w : int
            Window/Grid width.

        Returns:
        -------
        torch.Tensor
            relative_coords (torch.Tensor): Pair-wise relative position indexes
            [height * width, height * width].
        """
        coords = torch.stack(torch.meshgrid([torch.arange(win_h), torch.arange(win_w)]))
        coords_flatten = torch.flatten(coords, 1)
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]
        relative_coords = relative_coords.permute(1, 2, 0).contiguous()
        relative_coords[:, :, 0] += win_h - 1
        relative_coords[:, :, 1] += win_w - 1
        relative_coords[:, :, 0] *= 2 * win_w - 1
        return relative_coords.sum(-1)

    def _get_relative_positional_bias(self) -> torch.Tensor:
        """
        Returns the relative positional bias.

        Returns:
        -------
        torch.Tensor
            relative_position_bias (torch.Tensor): Relative positional bias.
        """
        relative_position_bias = self.relative_position_bias_table[
            self.relative_position_index.view(-1)
        ].view(self.attn_area, self.attn_area, -1)
        relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous()
        return relative_position_bias.unsqueeze(0)

    def forward(self) -> torch.Tensor:
        """
        Forward Method

        Returns:
        -------
        torch.Tensor
            Pairwise relative position bias
        """
        return self._get_relative_positional_bias()



"""
Implementation of MaxViT module
"""
from dataclasses import dataclass
from typing import List, Tuple, Type, Union

import torch
from torch import nn

from metnet.layers.MBConv import MBConv
from metnet.layers.PartitionAttention import BlockAttention, GridAttention


@dataclass
class MaxViTDataClass:
    """
    DataClass for MaxViT

    Parameters
    ----------
    mb_conv_expansion_rate : int, optional
        MBConv: Expansion rate for the output channels, by default 4
    mb_conv_downscale : bool, optional
        MBConv: Flag to denote downscaling in the conv branch, by default False
    mb_conv_act_layer : Type[nn.Module], optional
        MBConv: activation layer, by default nn.GELU
    mb_conv_norm_layer : Type[nn.Module], optional
        MBConv: norm layer, by default nn.BatchNorm2D
    mb_conv_drop_path : float, optional
        MBConv: Stochastic Depth ratio, by default 0.0
    mb_conv_kernel_size : int, optional
        MBConv: Conv kernel size, by default 3
    mb_conv_se_bottleneck_ratio : float, optional
        MBConv: Squeeze Excite reduction ratio, by default 0.25
    block_attention_num_heads : int, optional
        BlockAttention: Number of attention heads, by default 32
    block_attention_channels : int
        BlockAttention: Number of channels used for attention computations.
        It should be divisible by num_heads, by default 64
    block_attention_attn_grid_window_size : Tuple[int, int], optional
        BlockAttention: Grid/Window size for attention, by default (8, 8)
    block_attention_attn_drop : float, optional
        BlockAttention: Dropout ratio of attention weight, by default 0
    block_attention_proj_drop : float, optional
        BlockAttention: Dropout ratio of output, by default 0
    block_attention_drop_path : float, optional
        BlockAttention: Stochastic depth, by default 0
    block_attention_pre_norm_layer : Type[nn.Module], optional
        BlockAttention: Pre norm layer, by default nn.LayerNorm
    block_attention_post_norm_layer : Type[nn.Module], optional
        BlockAttention: Post norm layer, by default nn.LayerNorm
    block_attention_use_mlp : Type[nn.Module], optional
        BlockAttention: MLP to be used after the attention, by default None
    block_attention_use_normalised_qk : bool, optional
        BlockAttention: Normalise queries and keys as done in Metnet 3, by default True
    grid_attention_num_heads : int, optional
        GridAttention: Number of attention heads, by default 32
    grid_attention_channels : int
        GridAttention: Number of channels used for attention computations.
        It should be divisible by num_heads, by default 64
    grid_attention_attn_grid_window_size : Tuple[int, int], optional
        GridAttention: Grid/Window size for attention, by default (8, 8)
    grid_attention_attn_drop : float, optional
        GridAttention: Dropout ratio of attention weight, by default 0
    grid_attention_proj_drop : float, optional
        GridAttention: Dropout ratio of output, by default 0
    grid_attention_drop_path : float, optional
        GridAttention: Stochastic depth, by default 0
    grid_attention_pre_norm_layer : Type[nn.Module], optional
        GridAttention: Pre norm layer, by default nn.LayerNorm
    grid_attention_post_norm_layer : Type[nn.Module], optional
        GridAttention: Post norm layer, by default nn.LayerNorm
    grid_attention_use_mlp : Type[nn.Module], optional
        GridAttention: MLP to be used after the attention, by default None
    grid_attention_use_normalised_qk : bool, optional
        GridAttention: Normalise queries and keys as done in Metnet 3, by default True
    """

    mb_conv_expansion_rate: int = 4
    mb_conv_downscale: bool = False
    mb_conv_act_layer: Type[nn.Module] = nn.GELU
    mb_conv_norm_layer: Type[nn.Module] = nn.BatchNorm2d
    mb_conv_drop_path: float = 0.0
    mb_conv_kernel_size: int = 3
    mb_conv_se_bottleneck_ratio: float = 0.25
    block_attention_num_heads: int = 32
    block_attention_channels: int = 64
    block_attention_attn_grid_window_size: Tuple[int, int] = (8, 8)
    block_attention_attn_drop: float = 0
    block_attention_proj_drop: float = 0
    block_attention_drop_path: float = 0
    block_attention_pre_norm_layer: Type[nn.Module] = nn.LayerNorm
    block_attention_post_norm_layer: Type[nn.Module] = nn.LayerNorm
    block_attention_use_mlp: Type[nn.Module] = None
    block_attention_use_normalised_qk: bool = True
    grid_attention_num_heads: int = 32
    grid_attention_channels: int = 64
    grid_attention_attn_grid_window_size: Tuple[int, int] = (8, 8)
    grid_attention_attn_drop: float = 0
    grid_attention_proj_drop: float = 0
    grid_attention_drop_path: float = 0
    grid_attention_pre_norm_layer: Type[nn.Module] = nn.LayerNorm
    grid_attention_post_norm_layer: Type[nn.Module] = nn.LayerNorm
    grid_attention_use_mlp: Type[nn.Module] = None
    grid_attention_use_normalised_qk: bool = True


class MaxViTBlock(nn.Module):
    """
    MaxViT block
    """

    def __init__(self, in_channels: int, maxvit_config: Type[MaxViTDataClass]) -> None:
        """
        Constructor Method

        Parameters
        ----------
        in_channels : int
            Number of input channels
        maxvit_config : Type[MaxViTDataClass]
            MaxVit Config
        """
        super().__init__()
        self.in_channels = in_channels
        self.config = maxvit_config

        self.mb_conv = MBConv(
            in_channels=self.in_channels,
            expansion_rate=self.config.mb_conv_expansion_rate,
            downscale=self.config.mb_conv_downscale,
            act_layer=self.config.mb_conv_act_layer,
            norm_layer=self.config.mb_conv_norm_layer,
            drop_path=self.config.mb_conv_drop_path,
            kernel_size=self.config.mb_conv_kernel_size,
            se_bottleneck_ratio=self.config.mb_conv_se_bottleneck_ratio,
        )

        # Init Block and Grid Attention

        self.block_attention = BlockAttention(
            in_channels=self.in_channels,
            num_heads=self.config.block_attention_num_heads,
            attention_channels=self.config.block_attention_channels,
            attn_grid_window_size=self.config.block_attention_attn_grid_window_size,
            attn_drop=self.config.block_attention_attn_drop,
            proj_drop=self.config.block_attention_proj_drop,
            drop_path=self.config.block_attention_drop_path,
            pre_norm_layer=self.config.block_attention_pre_norm_layer,
            post_norm_layer=self.config.block_attention_post_norm_layer,
            use_mlp=self.config.block_attention_use_mlp,
            use_normalised_qk=self.config.block_attention_use_normalised_qk,
        )

        self.grid_attention = GridAttention(
            in_channels=self.in_channels,
            num_heads=self.config.grid_attention_num_heads,
            attention_channels=self.config.grid_attention_channels,
            attn_grid_window_size=self.config.grid_attention_attn_grid_window_size,
            attn_drop=self.config.grid_attention_attn_drop,
            proj_drop=self.config.grid_attention_proj_drop,
            drop_path=self.config.grid_attention_drop_path,
            pre_norm_layer=self.config.grid_attention_pre_norm_layer,
            post_norm_layer=self.config.grid_attention_post_norm_layer,
            use_mlp=self.config.grid_attention_use_mlp,
            use_normalised_qk=self.config.grid_attention_use_normalised_qk,
        )

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """
        Forward pass.

        Parameters
        ----------
        X : torch.Tensor
            Input tensor of the shape [N, C, H, W]

        Returns:
        -------
        torch.Tensor
            MaxViT block output tensor of the shape [N, C, H, W]
        """

        output = self.mb_conv(X)
        output = self.block_attention(output)
        output = self.grid_attention(output)

        return output


class MetNetMaxVit(nn.Module):
    """
    MaxViT block implemented with MetNet 3 modifications.
    """

    def __init__(
        self,
        in_channels: int = 512,
        out_channels: int = None,
        num_blocks: int = 12,
        maxvit_conf: Union[Type[MaxViTDataClass], List[Type[MaxViTDataClass]]] = MaxViTDataClass(),
        set_linear_stocastic_depth: bool = True,
    ) -> None:
        """
        MetNet3 MaxViT Block

        Parameters
        ----------
        in_channels : int, optional
            Input Channels, by default 512
        out_channels : int, optional
            Output Channels, by default None
        num_blocks : int, optional
            Number of MaxViT blocks, by default 12
        maxvit_conf : Union[ Type[MaxViTDataClass], List[Type[MaxViTDataClass]] ], optional
            MaxViT config, by default MaxViTDataClass()
        set_linear_stocastic_depth : bool, optional
            Flag to set the stochastic depth linearly in each MaxVit subblock, by default True
        """
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels if out_channels else in_channels

        self.num_blocks = num_blocks
        self.set_linear_stocastic_depth = set_linear_stocastic_depth

        self.maxvit_blocks = nn.ModuleList()

        if isinstance(maxvit_conf, List):
            assert len(maxvit_conf) == num_blocks
            self.maxvit_conf_list = maxvit_conf
        else:
            self.maxvit_conf_list = [maxvit_conf for _ in range(self.num_blocks)]

        if self.set_linear_stocastic_depth:
            # Linearly sets the stochastic depth a given sub-module
            # (i.e. MBConv, local (block) attention or gridded (grid) attention)
            # from 0 to 0.2, as mentioned in Metnet3 paper
            for conf in self.maxvit_conf_list:
                conf.mb_conv_drop_path = 0
                conf.block_attention_drop_path = 0.1
                conf.grid_attention_drop_path = 0.2

        for conf in self.maxvit_conf_list:
            self.maxvit_blocks.append(MaxViTBlock(in_channels=self.in_channels, maxvit_config=conf))

        self.linear_transform = nn.Conv2d(
            in_channels=self.in_channels, out_channels=self.in_channels, kernel_size=1
        )

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """
        Forward method

        Parameters
        ----------
        X : torch.Tensor
            Input tensor

        Returns:
        -------
        torch.Tensor
            Output of the MaxViT block.
        """
        model_output_list = []
        model_output_list.append(self.maxvit_blocks[0](X))
        for i in range(1, self.num_blocks):
            model_output_list.append(self.maxvit_blocks[i](model_output_list[i - 1]))

        output = X + torch.stack(model_output_list).sum(dim=0)

        output = self.linear_transform(output)
        return output
